// XCamera.cpp: implementation of the XCamera class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#define		DISPLAY_API	__declspec(dllexport)

#include "../common/common.h"
#include "interfaces.h"
#include "manager.h"
#include "XCamera.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

XCamera::XCamera()
{
	m_lpDevice = NULL;
	m_lpCameraInfo = NULL;

	m_fZoomDistance = 50.0f;	

	m_vLookAtPoint = D3DXVECTOR3(10, 10, 10); // origin	
	m_fViewAngle = -D3DX_PI * 3 / 4; // 135 degree
	
	m_fScrlVelX = 0.0f;
	m_fScrlVelY = 0.0f;
	m_fZAngleVel = 0.0f;
	m_fRotVel = 0.0f;
	m_fZoomVel = 0.0f;

	m_iMinPos = 0;
	m_iMaxPos = 10;
	m_fZAngle = Z_ANGLE;
	m_dwRef = 0;

	m_fCameraHeight = 0.0f;
}

XCamera::~XCamera()
{
	m_lpDevice = NULL;
	SAFE_RELEASE(m_lpCameraInfo);
}

HRESULT		XCamera::AddRef()
{
	return m_dwRef++;
}

HRESULT		XCamera::Release()
{
	if(m_dwRef > 0)
	{
		m_dwRef--;
	}
	if(m_dwRef == 0)
	{			
		delete this;
		TRACE("XCamera object destroyed.");
		return 0;
	}	
	else
	{
		return m_dwRef;
	}	
}
//
// Inicialization
HRESULT		XCamera::InitCamera(int iMinBound, int iMaxBound)
{
	HRESULT dwRet = S_FALSE;
	IDisplay *dis = NULL;
	CreateDisplayObject(DISIID_IDisplay, (void**) &dis);
	if(!dis)
	{
		XException exp("LoadTextureFromFile: Device is invalid.");
		THROW(exp);			
	}
	// Get device from IDisplay
	m_lpDevice = dis->GetDevice();
	SAFE_RELEASE(dis);
	if(m_lpDevice)
	{		
		m_iMinPos = iMinBound;
		m_iMaxPos = iMaxBound; 

		CreateDisplayObject(DISIID_I3DFont, (void**)&m_lpCameraInfo);
		m_lpCameraInfo->CreateFont(12, 6, "Arial", FALSE, D3DCOLOR_ARGB(255,220,220,255));
		//
		// Set transformations	
		D3DXMATRIX matWorld;		
		//
		// Set world matrix
		D3DXMatrixIdentity(&matWorld);
		m_lpDevice->SetTransform(D3DTS_WORLD, &matWorld);	
		//
		// Create view matrix
		D3DXVECTOR3	vUpDir;
		m_vEyePt.x = m_fZoomDistance * cosf(m_fViewAngle) + m_vLookAtPoint.x;
		m_vEyePt.y = m_fZoomDistance * sinf(m_fViewAngle) + m_vLookAtPoint.y;
		m_vEyePt.z = m_fZoomDistance * sinf(Z_ANGLE);
		//
		vUpDir = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
		//
		D3DXMatrixLookAtLH(&m_matView, &m_vEyePt, &m_vLookAtPoint, &vUpDir);	
		D3DXMatrixPerspectiveFovLH(&m_matProjection, FOV, ASPECT, NEAR_CLIP, FAR_CLIP);	
		
		m_lpDevice->SetTransform(D3DTS_VIEW, &m_matView);
		m_lpDevice->SetTransform(D3DTS_PROJECTION, &m_matProjection);		
		dwRet = S_OK;
	}
	return dwRet;
}

// Restores camera matrices
HRESULT		XCamera::RestoreCamera()
{
	D3DXMATRIX matWorld;		
	if(m_lpDevice)
	{	
		//
		// Set world matrix
		D3DXMatrixIdentity(&matWorld);
		m_lpDevice->SetTransform(D3DTS_WORLD, &matWorld);
		m_lpDevice->SetTransform(D3DTS_VIEW, &m_matView);
		m_lpDevice->SetTransform(D3DTS_PROJECTION, &m_matProjection);		
		return S_OK;
	}
	return S_FALSE;
}


// Update
HRESULT		XCamera::ProcessCamera(float	fElapsedTime)
{
	DWORD dwRet = S_FALSE; // CAMERA NOT INITIALIZED
	if(m_lpDevice) {		
	
		D3DXVECTOR3	vUpDir;
		vUpDir = D3DXVECTOR3(0.0f, 0.0f, 1.0f);
		//
		// Apply rotation
		m_fViewAngle += m_fRotVel * fElapsedTime;
		//
		// Apply zooming
		m_fZoomDistance += m_fZoomVel * fElapsedTime;
		//
		// Check zoom distance
		if(m_fZoomDistance > MAX_ZOOM)
			m_fZoomDistance = MAX_ZOOM;
		if(m_fZoomDistance < MIN_ZOOM)
			m_fZoomDistance = MIN_ZOOM;
		//
		// Apply Z angle
		m_fZAngle += m_fZAngleVel * fElapsedTime;		
		if(m_fZAngle > MAX_Z_ANGLE)
			m_fZAngle = MAX_Z_ANGLE;
		if(m_fZAngle < MIN_Z_ANGLE)
			m_fZAngle = MIN_Z_ANGLE;
		//	
		// Apply position
		m_vLookAtPoint.x += (m_fScrlVelX * cosf(m_fViewAngle + D3DX_PI / 2)) * fElapsedTime;
		m_vLookAtPoint.y += (m_fScrlVelX * sinf(m_fViewAngle + D3DX_PI / 2)) * fElapsedTime;
		m_vLookAtPoint.x += (m_fScrlVelY * cosf(m_fViewAngle)) * fElapsedTime;
		m_vLookAtPoint.y += (m_fScrlVelY * sinf(m_fViewAngle)) * fElapsedTime;
		m_vLookAtPoint.z = m_fCameraHeight;
		//
		// Bound camera position
		if(m_vLookAtPoint.x < m_iMinPos) 
			m_vLookAtPoint.x = (float)m_iMinPos;
		if(m_vLookAtPoint.y < m_iMinPos) 
			m_vLookAtPoint.y = (float)m_iMinPos;
		if(m_vLookAtPoint.x > m_iMaxPos) 
			m_vLookAtPoint.x = (float)m_iMaxPos;
		if(m_vLookAtPoint.y > m_iMaxPos) 
			m_vLookAtPoint.y = (float)m_iMaxPos;		
		//
		// Computes eye point position
		m_vEyePt.x = m_fZoomDistance * cosf(m_fViewAngle) + m_vLookAtPoint.x;
		m_vEyePt.y = m_fZoomDistance * sinf(m_fViewAngle) + m_vLookAtPoint.y;
		m_vEyePt.z = m_fZoomDistance * sinf(m_fZAngle) + m_fCameraHeight;		
		//
		// Create new view matrix
		D3DXMatrixLookAtLH(&m_matView, &m_vEyePt, &m_vLookAtPoint, &vUpDir);
		// Apply transfromation		
		m_lpDevice->SetTransform(D3DTS_VIEW, &m_matView);
		//
		// Reset velocities
		m_fScrlVelX = 0;
		m_fScrlVelY = 0;
		m_fRotVel = 0;
		m_fZoomVel = 0;
		m_fZAngleVel = 0;
				
		dwRet = ERROR_SUCCESS;
		//
		// Computes billboard matrix
		D3DXVECTOR3 vDir = m_vLookAtPoint - m_vEyePt;
		if( vDir.y > 0.0f ) {
			D3DXMatrixRotationZ(&m_matBillboard, -atanf(vDir.x/vDir.y)+D3DX_PI/2 );
		}
		else {
			D3DXMatrixRotationZ(&m_matBillboard, -atanf(vDir.x/vDir.y)-D3DX_PI/2 );
		}			
		static char szCameraInfo[128];
		sprintf(szCameraInfo, "LookAtPoint: %2.1lf, %2.1lf, %2.1lf\nEyePoint: %2.1lf, %2.1lf, %2.1lf", m_vLookAtPoint.x, m_vLookAtPoint.y, m_vLookAtPoint.z, m_vEyePt.x, m_vEyePt.y, m_vEyePt.z);
		m_lpCameraInfo->Draw(szCameraInfo, 0, 50);				
	}
	return dwRet;

}
