//*****************************************************************************
//
//    System          : jmsAPI
//    Category        : JMS exception classes
//    Source          : JMS API documentation 
//
//    Copyright (c), Leonov Alexander
//
//    File version $Revision:1.0$ at $Date:24.11.2003 13:40:34$
//
//    Author : Leonov Alexander
//    web    : http://home.ural.ru/~tigra/
//    e-mail : shura@ol405.paco.net
//
//*****************************************************************************

unit jmsException;

interface

uses
  SysUtils, Classes;

type
  {{
  This is the root class of all JMS API exceptions. 
  
  It provides the following information: 
  
  A provider-specific string describing the error. This string is the standard 
  exception message and is available via the getMessage method. 
  A provider-specific string error code 
  A reference to another exception. Often a JMS API exception will be the 
  result of a lower-level problem. If appropriate, this lower-level exception 
  can be linked to the JMS API exception. 
  }
  EJMSException = class (Exception)
  protected
    FErrorCode: string;
    FLinkedException: Exception;
    function getErrorCode: string;
    function getLinkedException: Exception;
    procedure setLinkedException(ex: Exception);
  public
    {{
    Constructs a JMSException with the specified reason and with the error code 
    defaulting to null.
    Parameters:
    reason - a description of the exception
    }
    constructor Create(reason : string); overload;
    {{
    Constructs a JMSException with the specified reason and error code.
    Parameters:
    reason - a description of the exception
    errorCode - a string specifying the vendor-specific error code
    }
    constructor Create(reason : string; errorCode : string); overload;
    {{
    Gets the vendor-specific error code.
    Returns:
    a string specifying the vendor-specific error code
    }
    property ErrorCode: string read getErrorCode;
    {{
    Gets/Set the exception linked to this one.
    }
    property LinkedException: Exception read getLinkedException write 
            setLinkedException;
  end;
  
  {{
  This exception is thrown when a method is invoked at an illegal or 
  inappropriate time or if the provider is not in an appropriate state for the 
  requested operation. For example, this exception must be thrown if 
  Session.commit is called on a non-transacted session. 
  }
  EIllegalStateException = class (EJMSException)
  end;
  
  {{
  This exception must be thrown when a client attempts to set a connection's 
  client ID to a value that is rejected by a provider. 
  }
  EInvalidClientIDException = class (EJMSException)
  end;
  
  {{
  This exception must be thrown when a destination either is not understood by 
  a provider or is no longer valid. 
  }
  EInvalidDestinationException = class (EJMSException)
  end;
  
  {{
  This exception must be thrown when a JMS client attempts to give a provider a 
  message selector with invalid syntax. 
  }
  EInvalidSelectorException = class (EJMSException)
  end;
  
  {{
  This exception must be thrown when a provider rejects a user name/password 
  submitted by a client. It may also be thrown for any case where a security 
  restriction prevents a method from completing. 
  }
  EJMSSecurityException = class (EJMSException)
  end;
  
  {{
  This exception must be thrown when an unexpected end of stream has been 
  reached when a StreamMessage or BytesMessage is being read. 
  }
  EMessageEOFException = class (EJMSException)
  end;
  
  {{
  This exception must be thrown when a JMS client attempts to use a data type 
  not supported by a message or attempts to read data in a message as the wrong 
  type. It must also be thrown when equivalent type errors are made with 
  message property values. For example, this exception must be thrown if 
  StreamMessage.writeObject is given an unsupported class or if 
  StreamMessage.readShort is used to read a boolean value. Note that the 
  special case of a failure caused by an attempt to read improperly formatted 
  String data as numeric values must throw the java.lang.NumberFormatException
  }
  EMessageFormatException = class (EJMSException)
  end;
  
  {{
  This exception must be thrown when a JMS client attempts to read a write-only 
  message. 
  }
  EMessageNotReadableException = class (EJMSException)
  end;
  
  {{
  This exception must be thrown when a JMS client attempts to write to a 
  read-only message. 
  }
  EMessageNotWriteableException = class (EJMSException)
  end;
  
  {{
  This exception is thrown when a provider is unable to allocate the resources 
  required by a method. For example, this exception should be thrown when a 
  call to TopicConnectionFactory.createTopicConnection fails due to a lack of 
  JMS provider resources. 
  }
  EResourceAllocationException = class (EJMSException)
  end;
  
  {{
  This exception is thrown when an operation is invalid because a transaction 
  is in progress. For instance, an attempt to call Session.commit when a 
  session is part of a distributed transaction should throw a 
  TransactionInProgressException. 
  }
  ETransactionInProgressException = class (EJMSException)
  end;
  
  {{
  This exception must be thrown when a call to Session.commit results in a 
  rollback of the current transaction. 
  }
  ETransactionRolledBackException = class (EJMSException)
  end;
  

implementation

{
******************************** EJMSException *********************************
}
constructor EJMSException.Create(reason : string);
begin
  Create (reason, '');
end;{ EJMSException.Create }

constructor EJMSException.Create(reason : string; errorCode : string);
begin
  FErrorCode := errorCode;
  inherited Create (reason);
end;{ EJMSException.Create }

function EJMSException.getErrorCode: string;
begin
  Result := FErrorCode;
end;{ EJMSException.getErrorCode }

function EJMSException.getLinkedException: Exception;
begin
  Result := FLinkedException;
end;{ EJMSException.getLinkedException }

procedure EJMSException.setLinkedException(ex: Exception);
begin
  FLinkedException := ex;
end;{ EJMSException.setLinkedException }


end.
