{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpIdrisi.pas. The Original Code is available through   }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit imports spatial data from Idrisi raster files. A header file with  }
{ the same name as the data file but with an .rdc extension must be present as }
{ well. (Currently coordinate system information is not parsed and a local     }
{ coordinate system is assigned.)                                              }
{ This driver currently only supports non-compressed binary data, RGB8 and     }
{ RGB24. Other formats (ASCII and compressed data) will be supported in a later}
{ version of this driver. RGB8 and RGB24 data are transformed into rasters of  }
{ TDSpUInt8 data type with 1 or 3 bands, respectively.                         }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 24                                             }
{                                                                              }
{ Dependencies  : DSpPFA                                                       }
{                 DSpUtils (SwapXBytes)                                        }
{                 DSpGlobals                                                   }
{                 Classes (TFileStream)                                        }
{                 SysUtils (Exception, ChangeFileExt, ExtractFileExt)          }
{                                                                              }
{ Version       : 0.1  03Nov24 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpIdrisi;
{$I DSpatial.inc}
interface
uses DSpGlobals, DSpPFA;
type
  TDSpIdrisiDriver = class(TDSpBaseDriver)
  private
    FFileName : string;
    FTitle    : string;
    FBands, FCols, FRows: integer;
    FDataType : TDSpDataType;
    FBandType : TDSpRasterBandType;
    FPixelSize: TDSpFloat64;
    FNull     : TDSpFloat64;
    FULX, FULY: TDSpFloat64;
    FMin, FMax: string; // there could be one or three values
  protected
    procedure   SetFileName(const fn: string);
    procedure   ReadRDC;

    property    FileName: string read FFileName write SetFileName;
    function    Description: string; override;
    function    DataSets: integer; override;

    function    DataSetID(i: integer): integer; override;
    function    DataSetDescription(i: integer): string; override;
    function    DataSetCoordSys(i: integer): string; override;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); override;
    function    DataSetType(i: integer): TDSpPFAFormat; override;
  public
    constructor Create;

    procedure   AddRasterDataSet(const r: TDSpSourceDataSet); override;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); override;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); override;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); override;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); override;
  end;

  TDSpIdrisiBroker = class(TDSpBaseBroker)
  public
    constructor Create;
    function    CreateSourcehandler(const fn: string): TDSpSourceHandler; override;
  end;

implementation
uses SysUtils, Classes, DSpUtils;

////////////////////////////////////////////////////////////////////////////////
// TDSpIdrisiDriver
constructor TDSpIdrisiDriver.Create;
begin
  inherited;
  FBands := 1;
  FBandType := dspRasterBandOrdinal;
  FPixelSize := 1.0;
  FNull := -9E200;             // just an improbable value
end;

procedure TDSpIdrisiDriver.SetFileName(const fn: string);
var ext: string;
begin
  FFileName := fn;

  // Read the header file
  ext := UpperCase(ExtractFileExt(FFileName));
  ReadRDC;

  // Read the projection file

end;

procedure TDSpIdrisiDriver.ReadRDC;
var hdr: TextFile;
    line, token, value: string;
    white: integer;
begin
  AssignFile(hdr, ChangeFileExt(FFileName, '.RDC'));
  try
    Reset(hdr);
    while not Eof(hdr) do begin
      ReadLn(hdr, line);
      white := Pos(':', line);
      token := UpperCase(Trim(Copy(line, 1, white-1)));
      value := UpperCase(Trim(Copy(line, white+1, 1000)));
      if token = 'COLUMNS' then FCols := StrToInt(value)
      else if token = 'ROWS' then FRows := StrToInt(value)
      else if token = 'FILE TITLE' then FTitle := value
      else if token = 'MIN. X' then FULX := StrToFloat(value)
      else if token = 'MAX. Y' then FULY := STrToFloat(value)
      else if token = 'RESOLUTION' then FPixelSize := StrToFloat(value)
      else if token = 'FLAG VALUE' then FNull := StrToFloat(value)
      else if token = 'MIN. VALUE' then FMin := value
      else if token = 'MAX. VALUE' then FMax := value
      else if token = 'DATA TYPE' then begin
        if value = 'REAL' then FDataType := dspFloat32
        else if value = 'BYTE' then FDataType := dspUInt8
        else if value = 'INTEGER' then FDataType := dspInt16
        else if value = 'RGB8' then begin
          FDataType := dspUInt8;
          FBandType := dspRasterBandImage;
        end else if value = 'RGB24' then begin
          FDataType := dspUInt8;
          FBandType := dspRasterBandImage;
          FBands := 3;
        end;
      end;
    end;
  finally
    CloseFile(hdr);
  end;

  // Check for fatal combinations and omissions
  if (FCols = 0) or (FRows = 0) or FloatsSame(FPixelSize, 0) then begin
    ; // show parameter dialog
  end;
end;

function TDSpIdrisiDriver.Description: string;
begin
  if FTitle <> '' then Result := FTitle
  else Result := 'Idrisi raster file: ' + FFileName;
end;

function TDSpIdrisiDriver.DataSets: integer;
begin
  Result := 1;
end;

function TDSpIdrisiDriver.DataSetID(i: integer): integer;
begin
  Result := 0;
end;

function TDSpIdrisiDriver.DataSetDescription(i: integer): string;
begin
  if FTitle <> '' then Result := FTitle
  else Result := ExtractFileName(FFileName);
end;

function TDSpIdrisiDriver.DataSetCoordSys(i: integer): string;
begin
  Result := 'LOCAL_CS["Local coordinate system (meter)",LOCAL_DATUM["Local datum",0],UNIT["Meter",1.0],AXIS["X",EAST],AXIS["Y",NORTH]]';  // <<<<<<<<<<<< projection not yet supported
end;

procedure TDSpIdrisiDriver.DataSetExtent(i: integer; const ext: TDSpExtent);
begin
  ext.North := FULY;
  ext.West  := FULX;
  ext.East  := ext.West + FCols * FPixelSize;
  ext.South := ext.North - FRows * FPixelSize;
end;

function TDSpIdrisiDriver.DataSetType(i: integer): TDSpPFAFormat;
begin
  Result := dspPFARaster;
end;

procedure TDSpIdrisiDriver.AddRasterDataSet(const r: TDSpSourceDataSet);
begin
  Assert(false, 'Idrisi driver does not support writing of new data sets');
end;

procedure TDSpIdrisiDriver.GetRasterInfo(i: integer; info: PDSpRasterInfo);
begin
  if FBands = 1 then info^.BandFormat := dspRasterFlat
  else info^.BandFormat := dspRasterInterleavePixel;
  info^.Bands := FBands;
  info^.Rows := FRows;
  info^.Columns := FCols;
  info^.PixelSize := FPixelSize;
end;

procedure TDSpIdrisiDriver.GetRasterBandInfo(i: integer; info: PDSpRasterInfo);
var b, c: integer;
    minimum, maximum: array [0..2] of TDSpFloat64;
begin
  // Estiblish band minimum and maximum values
  if FBands = 1 then begin
    minimum[0] := StrToFloat(FMin);
    maximum[0] := StrToFloat(FMax);
  end else begin
    // RGB24, FMin and FMax hold three values each
    b := Pos(' ', FMin);
    c := Pos(' ', Copy(FMin, b+1, 80));
    minimum[0] := STrToFloat(Copy(FMin, 1, b-1));
    minimum[1] := STrToFloat(Copy(FMin, b+1, c-1));
    minimum[2] := STrToFloat(Copy(FMin, c+1, 80));
    b := Pos(' ', FMax);
    c := Pos(' ', Copy(FMax, b+1, 80));
    maximum[0] := STrToFloat(Copy(FMax, 1, b-1));
    maximum[1] := STrToFloat(Copy(FMax, b+1, c-1));
    maximum[2] := STrToFloat(Copy(FMax, c+1, 80));
  end;

  for b := 0 to High(info^.Band) do
    with info^.Band[b] do begin
      BandIndex := b;
      BandType := FBandType;
      DataType := FDataType;
      MissingValue := FNull;
      Min := minimum[b];
      Max := maximum[b];
    end;
end;

procedure TDSpIdrisiDriver.ReadRasterBand(dataSet, band: integer; const into: pointer);
// For now only reads flat binary and RGB8/24 files
var r, c: integer;
    data: TFileStream;
    p, pix: PByte;
    row: PByte;
begin
  Assert((dataSet = 0) and (band < FBands));

  data := TFileStream.Create(FFileName, fmOpenRead or fmShareDenyWrite);
  try
    if FBands = 1 then data.Read(into^, FRows * FCols * DataSize(FDataType))
    else begin
      // Read RGB24 data
      GetMem(row, FCols * 3);
      try
        p := into;
        for r := 0 to FRows - 1 do begin
          data.Read(row^, FCols * 3);
          pix := row;
          Inc(pix, band);
          for c := 0 to FCols - 1 do begin
            p^ := pix^;
            Inc(p);
            Inc(pix, 3);
          end;
        end;
      finally
        FreeMem(row);
      end;
    end;
  finally
    data.Free;
  end;
end;

procedure TDSpIdrisiDriver.ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

procedure TDSpIdrisiDriver.ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpIdrisiBroker
constructor TDSpIdrisiBroker.Create;
begin
  inherited;
  FDescription := 'Idrisi raster files';
  FFilter := '*.RST';
  FExtension := 'RST';
  FCaps := [dspReadRaster];
end;

function TDSpIdrisiBroker.CreateSourceHandler(const fn: string): TDSpSourceHandler;
var r: TDSpIdrisiDriver;
begin
  r := TDSpIdrisiDriver.Create;
  r.FileName := fn;
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

initialization
{ Enable this method for static loading of the broker. Please note that the    }
{ DSpPFA unit MUST be loaded before this unit. If the broker is to be loaded   }
{ dynamically, you will have to provide a DLL with this unit.                  }
  PFA.RegisterBroker(TDSpIdrisiBroker.Create);
end.
