{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRawBinary.pas. The Original Code is available through}
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit imports spatial data from raw binary files. Data files can be in   }
{ BIL, BIP, or BSQ format, or an ESRI binary raster file. DEM files are also   }
{ supported and considered to contain a single band with elevation data. A     }
{ header file with the same name as the data file but with an .hdr extension   }
{ must be present as well. The driver will look for an optional third file with}
{ a .prj extension as well, containing coordinate system information in either }
{ WKT or ArcInfo format. In the absence of this latter file a local coordinate }
{ system is assigned to the data set. (Currently coordinate system information }
{ is not parsed and a local coordinate system is assigned.)                    }
{ This driver supports 8, 16 and 32 bit data. Eight bit data is interpreted as }
{ dspUInt8, 16 and 32 bit as dspInt16/32. ESRI rasters are always single band  }
{ dspFloat32.                                                                  }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 24                                             }
{                                                                              }
{ Dependencies  : DSpPFA                                                       }
{                 DSpUtils (SwapXBytes)                                        }
{                 DSpGlobals                                                   }
{                 Classes (TFileStream)                                        }
{                 SysUtils (Exception, ChangeFileExt, ExtractFileExt)          }
{                                                                              }
{ Version       : 0.3  03Nov24 Added support for ESRI binary rasters.          }
{                 0.2  03Nov07 PFA overhaul                                    }
{                 0.1  03Aug24 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpRawBinary;
{$I DSpatial.inc}
interface
uses DSpGlobals, DSpPFA;
type
  TDSpRawDriver = class(TDSpBaseDriver)
  private
    FFileName: string;
    FOrder: char; // I(ntel) or M(otorola)
    FLayout: string[3]; // BIL, BIP, BSQ
    FBands, FCols, FRows: integer;
    FBits, FSkip: integer;
    FDataType: TDSpDataType;
    FBandRow, FTotalRow, FBandGap: integer;
    FPixelSize, FDY: TDSpFloat64;
    FNull: TDSpFloat64;
    FULX, FULY: TDSpFloat64;
  protected
    procedure   SetFileName(const fn: string);
    procedure   ReadHDR;
    procedure   ReadFLTHDR;

    property    FileName: string read FFileName write SetFileName;
    function    Description: string; override;
    function    DataSets: integer; override;

    function    DataSetID(i: integer): integer; override;
    function    DataSetDescription(i: integer): string; override;
    function    DataSetCoordSys(i: integer): string; override;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); override;
    function    DataSetType(i: integer): TDSpPFAFormat; override;
  public
    constructor Create;

    procedure   AddRasterDataSet(const r: TDSpSourceDataSet); override;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); override;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); override;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); override;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); override;
  end;

  TDSpRawBroker = class(TDSpBaseBroker)
  public
    constructor Create;
    function    CreateSourcehandler(const fn: string): TDSpSourceHandler; override;
    function    SupportsExtension(ext: string): boolean; override;
  end;

implementation
uses SysUtils, Classes, DSpUtils;

////////////////////////////////////////////////////////////////////////////////
// TDSpRawDriver
constructor TDSpRawDriver.Create;
begin
  inherited;
  FOrder := 'I';                // set defaults
  FLayout := 'BIL';
  FBands := 1;
  FBits := 8;
  FPixelSize := 1.0;
  FNull := -9E200;             // just an improbable value
end;

procedure TDSpRawDriver.SetFileName(const fn: string);
var ext: string;
begin
  FFileName := fn;

  // Read the header file
  ext := UpperCase(ExtractFileExt(FFileName));
  if ext = '.FLT' then ReadFLTHDR else ReadHDR;

  // Read the projection file

end;

procedure TDSpRawDriver.ReadHDR;
var hdr: TextFile;
    line, token, value: string;
    white: integer;
begin
  AssignFile(hdr, ChangeFileExt(FFileName, '.hdr'));
  try
    Reset(hdr);
    while not Eof(hdr) do begin
      ReadLn(hdr, line);
      white := Pos(' ', line);
      if white = 0 then white := Pos(#9, line);
      token := UpperCase(Copy(line, 1, white-1));
      value := UpperCase(Trim(Copy(line, white+1, 1000)));

      if token = 'NROWS' then FRows := StrToInt(value)
      else if token = 'NCOLS' then FCols := StrToInt(value)
      else if token = 'NBANDS' then FBands := StrToInt(value)
      else if token = 'NBITS' then FBits := StrToInt(value)
      else if token = 'BYTEORDER' then FOrder := value[1]
      else if token = 'LAYOUT' then FLayout := value
      else if token = 'SKIPBYTES' then FSkip := StrToInt(value)
      else if token = 'ULXMAP' then FULX := StrToFloat(value)
      else if token = 'ULYMAP' then FULY := StrToFloat(value)
      else if token = 'XDIM' then FPixelSize := StrToFloat(value)
      else if token = 'YDIM' then FDY := StrToFloat(value)
      else if token = 'BANDROWBYTES' then FBandRow := StrToInt(value)
      else if token = 'TOTALROWBYTES' then FTotalRow := StrToInt(value)
      else if token = 'BANDGAPBYTES' then FBandGap := StrToInt(value)
      else if token = 'NODATA' then FNull := StrToFloat(value);
    end;
  finally
    CloseFile(hdr);
  end;

  // Check for fatal combinations??? FPixelSize <> FDY???

  // Some derived and updated parameters
  if FBits <= 8 then FDataType := dspUInt8
  else if FBits = 16 then FDataType := dspInt16 else FDataType := dspInt32;
  FULY := FULY + 0.5 * FPixelSize;     // correct to edge of pixel
  FULX := FULX - 0.5 * FPixelSize;
end;

procedure TDSpRawDriver.ReadFLTHDR;
var hdr: TextFile;
    line, token, value: string;
    white: integer;
begin
  AssignFile(hdr, ChangeFileExt(FFileName, '.hdr'));
  try
    Reset(hdr);
    while not Eof(hdr) do begin
      ReadLn(hdr, line);
      white := Pos(' ', line);
      if white = 0 then white := Pos(#9, line);
      token := UpperCase(Copy(line, 1, white-1));
      value := UpperCase(Trim(Copy(line, white+1, 1000)));
      if token = 'NCOLS' then FCols := StrToInt(value)
      else if token = 'NROWS' then FRows := StrToInt(value)
      else if token = 'XLLCORNER' then FULX := StrToFloat(value)
      else if token = 'YLLCORNER' then FULY := STrToFloat(value)  // must correct
      else if token = 'CELLSIZE' then FPixelSize := StrToFloat(value)
      else if token = 'NODATA_VALUE' then FNull := StrToFloat(value)
      else if token = 'BYTEORDER' then
        if UpperCase(value) = 'LSBFIRST' then FOrder := 'I' else FOrder := 'M';
    end;
  finally
    CloseFile(hdr);
  end;

  // Check for fatal combinations and omissions
  if (FCols = 0) or (FRows = 0) or FloatsSame(FPixelSize, 0) then begin
    ; // show parameter dialog
  end else FULY := FULY + (FCols - 1) * FPixelSize;

  // Some fixed parameters
  FBits := 32;
  FDataType := dspFloat32;
end;

function TDSpRawDriver.Description: string;
begin
  Result := 'Raw binary file: ' + FFileName;
end;

function TDSpRawDriver.DataSets: integer;
begin
  Result := 1;
end;

function TDSpRawDriver.DataSetID(i: integer): integer;
begin
  Result := 0;
end;

function TDSpRawDriver.DataSetDescription(i: integer): string;
begin
  Result := ExtractFileName(FFileName);
end;

function TDSpRawDriver.DataSetCoordSys(i: integer): string;
begin
  Result := 'LOCAL_CS["Local coordinate system (meter)",LOCAL_DATUM["Local datum",0],UNIT["Meter",1.0],AXIS["X",EAST],AXIS["Y",NORTH]]';  // <<<<<<<<<<<< projection not yet supported
end;

procedure TDSpRawDriver.DataSetExtent(i: integer; const ext: TDSpExtent);
begin
  ext.North := FULY;
  ext.West  := FULX;
  ext.East  := ext.West + FCols * FPixelSize;
  ext.South := ext.North - FRows * FPixelSize;
end;

function TDSpRawDriver.DataSetType(i: integer): TDSpPFAFormat;
begin
  Result := dspPFARaster;
end;

procedure TDSpRawDriver.AddRasterDataSet(const r: TDSpSourceDataSet);
begin
  Assert(false, 'Raw binary driver does not support writing of new data sets');
end;

procedure TDSpRawDriver.GetRasterInfo(i: integer; info: PDSpRasterInfo);
begin
  if (FBands = 1) or (FLayout = 'BSQ') then info^.BandFormat := dspRasterFlat
  else if FLayout = 'BIL' then info^.BandFormat := dspRasterInterleaveLine
  else info^.BandFormat := dsprasterInterleavePixel;
  info^.Bands := FBands;
  info^.Rows := FRows;
  info^.Columns := FCols;
  info^.PixelSize := FPixelSize;
end;

procedure TDSpRawDriver.GetRasterBandInfo(i: integer; info: PDSpRasterInfo);
var b: integer;
begin
  for b := 0 to High(info^.Band) do
    with info^.Band[b] do begin
      BandIndex := b;
      if UpperCase(ExtractFileExt(FFileName)) = '.DEM' then BandType := dspRasterBandSurface
      else BandType := dspRasterBandOrdinal;
      DataType := FDataType;
      MissingValue := FNull;
      Min := 0;
      Max := -1;
    end;
end;

procedure TDSpRawDriver.ReadRasterBand(dataSet, band: integer; const into: pointer);
// For now only reads multiple byte sized data: 8, 16, 32 bit
var r, c, rowSize, len, skip: integer;
    data: TFileStream;
    p: pointer;
begin
  Assert((dataSet = 0) and (band < FBands));
  Assert(FBits >= 8);

  data := TFileStream.Create(FFileName, fmOpenRead or fmShareDenyWrite);
  len := FBits shr 3; // len is length in bytes of pixel DOES NOT WORK FOR BINARY OR 4-BIT DATA
  rowSize := FCols * len;
  p := into;
  try
    data.Seek(FSkip, soFromBeginning);
    if (FBands = 1) or (FLayout = 'BSQ') then begin
      // Flat file, just move the block
      if band > 0 then data.Seek(band * FRows * rowSize + FBandGap, soFromCurrent);
      data.Read(p^, FRows * rowSize);
    end else if FLayout = 'BIL' then begin
      // Read BIL data
      if FBandRow = 0 then FBandRow := rowSize;
      if FTotalRow > 0 then skip := FTotalRow - (FBands - 1) * FBandRow else skip := 0;
      if band > 0 then data.Seek(band * FBandRow, soFromCurrent);
      for r := 0 to FRows - 1 do begin
        if data.Read(p^, rowSize) <> rowSize then
          raise Exception.Create('Unexpected end-of-file encountered');
        data.Seek(skip, soFromCurrent);    // skip over the other bands
        p := pointer(integer(p) + rowSize);
      end;
    end else begin
      // Read BIP data
      if FTotalRow > 0 then skip := FTotalRow - FBands * FCols * (FBits shr 3) else skip := 0;
      rowSize := (FBands - 1) * len;  // rowSize is now size of interleaved pixel spacing
      if band > 0 then data.Seek(band * len, soFromCurrent);
      for r := 0 to FRows - 1 do begin
        for c := 0 to FCols - 1 do begin
          if data.Read(p^, len) <> len then
            raise Exception.Create('Unexpected end-of-file encountered');
          data.Seek(rowSize, soFromCurrent);
          p := pointer(integer(p) + len);
        end;
        data.Seek(skip, soFromCurrent);
      end;
    end;
  finally
    data.Free;
  end;

  if FOrder = 'M' then begin
    if FBits = 16 then Swap2Bytes(into^, FRows * FCols)
    else if FBits = 32 then Swap4Bytes(into^, FRows * FCols);
  end;
end;

procedure TDSpRawDriver.ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

procedure TDSpRawDriver.ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRawBroker
constructor TDSpRawBroker.Create;
begin
  inherited;
  FDescription := 'Raw binary files';
  FFilter := '*.BIL;*.BIP;*.BSQ;*.DEM;*.FLT';
  FExtension := 'BIL';  // not used -> SupportsExtension override
  FCaps := [dspReadRaster];
end;

function TDSpRawBroker.CreateSourceHandler(const fn: string): TDSpSourceHandler;
var r: TDSpRawDriver;
begin
  r := TDSpRawDriver.Create;
  r.FileName := fn;
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

function TDSpRawBroker.SupportsExtension(ext: string): boolean;
begin
  Result := (ext = 'BIL') or (ext = 'BIP') or (ext = 'BSQ') or (ext = 'DEM') or (ext = 'FLT');
end;

initialization
{ Enable this method for static loading of the broker. Please note that the    }
{ DSpPFA unit MUST be loaded before this unit. If the broker is to be loaded   }
{ dynamically, you will have to provide a DLL with this unit.                  }
  PFA.RegisterBroker(TDSpRawBroker.Create);
end.
