{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpTempRaster.pas. The Original Code is available       }
{ through the home page of DSpatial at http://dspatial.sourceforge.net.        }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This driver temporarily holds raster data until the data is saved or         }
{ discarded. For larger rasters a temporary file is used, which is a simple    }
{ flat binary file comparable to the BSQ format, but no header will be written }
{ for it. Since the driver is merely a conduit between a file and an instance  }
{ of TDSpRasterData the functionality of the driver only comes into play when  }
{ huge rasters are temporarily created. The file format is very similar to the }
{ raw binary format (with no header information available for it), with the    }
{ exception that all data types are supported and that files will be deleted   }
{ when the instance of the driver is destroyed. Header information and other   }
{ metadata are kept in memory for the life of the object. A temporary raster   }
{ can contain only a single raster, with an unlimited number of bands.         }
{                                                                              }
{ Currently data is not written to a file (there is no interface yet), so this }
{ driver is a true dummy. However, the TDSpRasterData object needs a driver for}
{ consistency. The functionality is not diminished, as memory management,      }
{ forcing TDSpRasterData to flush data to file, is not yet implemented.        }
{                                                                              }
{ NOTE CAREFULLY: This driver is not optional like other drivers for data      }
{ formats. The kernel needs to be able to write temporary rasters in order to  }
{ manage memory usage of raster data sets. Temporary rasters may be created by }
{ analysis functions as intermediate results, so even if you do not explicitly }
{ create temporary rasters the kernel might still need them. In the PFA        }
{ instance the broker is loaded as the first static broker, independent of     }
{ where this unit is listed in the uses clause.                                }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 18                                             }
{                                                                              }
{ Dependencies  : DSpPFA                                                       }
{                 DSpRasterSource                                              }
{                 DSpGlobals                                                   }
{                 SysUtils (Exception)                                         }
{                                                                              }
{ Version       : 0.1  03Nov18 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpTempRaster;
{$I DSpatial.inc}
interface
uses DSpGlobals, DSpPFA;
type
  TDSpTempRasterDriver = class(TDSpBaseDriver)
  private
    FFileName: string;
    FFileWritten: boolean;
    FCols, FRows: integer;
    FPixelSize: TDSpFloat64;
    FPrecision: array of TDSpDataType;
    FBandType : array of TDSpRasterBandType;
    FNull: array of TDSpFloat64;
    FNorth, FEast, FSouth, FWest: TDSpFloat64;
  protected
    property    FileName: string read FFileName write FFileName;
    function    Description: string; override;
    function    DataSets: integer; override;

    function    DataSetID(i: integer): integer; override;
    function    DataSetDescription(i: integer): string; override;
    function    DataSetCoordSys(i: integer): string; override;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); override;
    function    DataSetType(i: integer): TDSpPFAFormat; override;
  public
    constructor Create(const fn: string);
    destructor  Destroy; override;

    procedure   AddRasterDataSet(const r: TDSpSourceDataSet); override;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); override;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); override;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); override;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); override;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); override;
  end;

  TDSpTempRasterBroker = class(TDSpBaseBroker)
  public
    constructor Create;
    function    CreateSourceHandler(const fn: string): TDSpSourceHandler; override;
  end;

implementation
uses SysUtils, DSpRasterSource;

////////////////////////////////////////////////////////////////////////////////
// TDSpTempRasterDriver
constructor TDSpTempRasterDriver.Create(const fn: string);
begin
  inherited Create;
  FFileName := fn;
  FFileWritten := false;
end;

destructor TDSpTempRasterDriver.Destroy;
begin
  if FileExists(FFileName) then DeleteFile(FFileName);
  Finalize(FPrecision);
  Finalize(FBandType);
  Finalize(FNull);
  inherited;
end;

procedure TDSpTempRasterDriver.AddRasterDataSet(const r: TDSpSourceDataSet);
var b: integer;
begin
  with r as TDSpRasterSource do begin
    FCols := Columns;
    FRows := Rows;
    FPixelSize := PixelSize;
    FNorth := Extent.North;
    FEast := Extent.East;
    FSouth := Extent.South;
    FWest := Extent.West;
    if Bands > 0 then begin
      SetLength(FBandType, Bands);
      SetLength(FPrecision, Bands);
      SetLength(FNull, Bands);
      for b := 0 to Bands - 1 do begin
        FBandType[b] := Band[b].BandType;
        FPrecision[b] := Band[b].DataType;
        FNull[b] := MissingValue[b];
      end;
    end;
  end;
end;

function TDSpTempRasterDriver.Description: string;
begin
  Result := 'Temporary file';
end;

function TDSpTempRasterDriver.DataSets: integer;
begin
  Result := 1;
end;

function TDSpTempRasterDriver.DataSetID(i: integer): integer;
begin
  Result := 0;
end;

function TDSpTempRasterDriver.DataSetDescription(i: integer): string;
begin
  Result := 'Temporary data set';
end;

function TDSpTempRasterDriver.DataSetCoordSys(i: integer): string;
begin
  Result := 'LOCAL_CS["Local coordinate system (meter)",LOCAL_DATUM["Local datum",0],UNIT["Meter",1.0],AXIS["X",EAST],AXIS["Y",NORTH]]';  // <<<<<<<<<<<< projection not yet supported
end;

procedure TDSpTempRasterDriver.DataSetExtent(i: integer; const ext: TDSpExtent);
begin
  ext.North := FNorth;
  ext.East := FEast;
  ext.South := FSouth;
  ext.West := FWest;
end;

function TDSpTempRasterDriver.DataSetType(i: integer): TDSpPFAFormat;
begin
  Result := dspPFARaster;
end;

procedure TDSpTempRasterDriver.GetRasterInfo(i: integer; info: PDSpRasterInfo);
begin
  info^.BandFormat := dspRasterFlat;
  info^.Bands := Length(FBandType);
  info^.Rows := FRows;
  info^.Columns := FCols;
  info^.PixelSize := FPixelSize;
end;

procedure TDSpTempRasterDriver.GetRasterBandInfo(i: integer; info: PDSpRasterInfo);
begin
  Assert((i >= 0) and (i < Length(FBandType)));
  with info^.Band[i] do begin
    BandIndex := i;
    DataType := FPrecision[i];
    BandType := FBandType[i];
    MissingValue := FNull[i];
    Min := 0;
    Max := -1;
  end;
end;

procedure TDSpTempRasterDriver.ReadRasterBand(dataSet, band: integer; const into: pointer);
var f: file;
    i, sum: integer;
begin
  Assert(dataSet = 0);
  // If FFileWritten is true, data has been written to a temporary file
  // before. In that case, read it. Otherwise, this driver knows of no data and
  // the calling TDSpRasterData instance has to make do with its allocated
  // memory unaltered after this method.
  if FFileWritten then begin
    // Read the data from the file
    AssignFile(f, FFileName);
    try
      Reset(f, 1);
      sum := 0;
      if band > 0 then begin  // hop over previous bands
        for i := 0 to band - 1 do sum := sum + DataSize(FPrecision[i]);
        Seek(f, sum * FRows * FCols);
      end;
      BlockRead(f, into^, FRows * FCols * DataSize(FPrecision[band]));
    finally
      CloseFile(f);
    end;
  end;
end;

procedure TDSpTempRasterDriver.ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

procedure TDSpTempRasterDriver.ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer);
begin
  Assert(false, 'not yet implemented');
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpTempRasterBroker
constructor TDSpTempRasterBroker.Create;
begin
  inherited;
  FDescription := 'Temporary files';
  FFilter := '*.BSQ';
  FExtension := 'BSQ';
  FCaps := []; // don't want to show this driver in any user dialogs
end;

function TDSpTempRasterBroker.CreateSourceHandler(const fn: string): TDSpSourceHandler;
var r: TDSpTempRasterDriver;
begin
  r := TDSpTempRasterDriver.Create(fn);
  FDrivers.Add(r);
  Result := TDSpSourceHandler.Create(r, fn);
end;

end.
