{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpPFA.pas. The Original Code is available through      }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit defines the Pluggable Format Architecture (PFA). The PFA interfaces}
{ between drivers that read and write the different formats of spatial data and}
{ the client application. Thus the application has a uniform view of disparate }
{ data access methods. The PFA has subclasses specific to the type of data     }
{ being accessed: vector, raster, tabular. In every application there should   }
{ be exactly one instance of the TDSpPFA class with global scope. That instance}
{ is created in the initialization section of this unit as the PFA variable. It}
{ is automatically destroyed in the finalization section upon closing of the   }
{ application. This unit must be listed in the uses clause of the project      }
{ before any of the data set classes and statically loaded data brokers, as    }
{ they might reference the PFA variable in their initialization clauses.       }
{                                                                              }
{ The PFA manages multiple data sources simultaneously, such that true multi-  }
{ processing can be achieved on multi-processor systems (every handler operates}
{ as a separate thread), but even on single processor systems the latency of   }
{ slower components such as hard disks and network access is offset through the}
{ separation by threads of slower code.                                        }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 27                                             }
{                                                                              }
{ Dependencies  : DSpDataMgr                                                   }
{                 DSpRasterSource                                              }
{                 DSpVectorSource                                              }
{                 DSpTempRaster                                                }
{                 DSpCoordinateSystem                                          }
{                 DSpUtils (ValidateIndex)                                     }
{                 DSpGlobals                                                   }
{                 SyncObjs (TEvent)                                            }
{                 SysUtils (ExtractFileExt, Exception,                         }
{                           TMultiReadExclusiveWriteSynchronizer)              }
{                 Classes (TThread)                                            }
{                                                                              }
{ Version       : 0.6  03Nov27 DSpPFA class uses synchronization lock. Added   }
{                              support for reading source data into memory.    }
{                              Source deletion now happens during idle time.   }
{                 0.5  03Nov07 PFA model overhaul                              }
{                 0.4  03Sep25 TDSpPFAObject now supports state flag and data  }
{                              set removal.                                    }
{                 0.3  03Aug14 Many changes to raster calls                    }
{                 0.2  03Jul18 Addition of TDSpPFAObject                       }
{                              Changes to TDSpPFAHandler                       }
{                 0.1  03Apr26 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpPFA;
{$I DSpatial.inc}
interface
uses Classes, SysUtils, SyncObjs, DSpGlobals, DSpUtils, DSpCoordinateSystem;

type
  TDSpBaseDriver    = class;
  TDSpSourceHandler = class;
  TDSpSourceDataSet = class;

  TDSpPFAMessage = record
                     DataSourceIndex : integer;
                     Event           : TEvent;
                     Msg             : cardinal;            // DSPMSG_*
                     Parameter       : pointer;
                   end;
  TDSpThreadMessage = ^TDSpPFAMessage;

  TDSpDataSetInfo = record
                      ID             : integer;       // identifier for use by the driver
                      DescriptiveName: string;
                      CoordSys       : string;
                      Extent         : TDSpExtent;
                      DataSetType    : TDSpPFAFormat;
                      DataSetData    : pointer;       // depends on data set type
                    end;

  // The Mother-of-all-Data sources. The class provides the basic structure of the
  // different kinds of geospatial data sources, including common meta-information
  // such as a unique identifier, projection parameters, extent, unit. Otherwise
  // this class is a mostly abstract ancestor to more useful classes.
  TDSpSourceDataSet = class
  protected
    FId         : TDSpUInt32;           // unique identifier
    FName       : string;               // descriptive name of the data set
    FRefs       : integer;              // number of references to this data set
    FLock       : TMultiReadExclusiveWriteSynchronizer;
    FIndex      : integer;
    FDriver     : TDSpBaseDriver;
    FDataSetType: TDSpPFAFormat;

    FExtent     : TDSpExtent;           // extreme coordinates of the data set
    FCoordSys   : TDSpCoordinateSystem; // coordinate system; never free a TDSpCoordinateSystem: reference counted

    procedure   ReadLock;
    procedure   WriteLock;
    procedure   StartWriting; virtual;
    procedure   EndWriting; virtual;

    procedure   PurgeData; virtual; abstract;
  public
    constructor Create(const driver: TDSpBaseDriver; dataSet: integer);
    constructor Open(const driver: TDSpBaseDriver; dataSet: integer);
    destructor  Destroy; override;
    procedure   Close; virtual; abstract;
    function    Clone(const driver: TDSpBaseDriver; dataSet: integer): TDSpSourceDataSet; virtual; abstract;

    procedure   AddRef;
    function    CloseRef: integer;
    procedure   ReadUnlock;
    procedure   WriteUnlock;

    procedure   SetCoordinateSystem(cs: TDSpCoordinateSystem); overload;
    procedure   SetCoordinateSystem(const wkt: string); overload;
    function    DescribeCS(descr: TDSpDescribe): string;
    function    CoordinateSystem: TDSpCoordinateSystem;
    procedure   SetExtent(ext: TDSpExtent);

    property    Id: TDSpUInt32 read FId;
    property    Name: string read FName write FName;
    property    DataSetType: TDSpPFAFormat read FDataSetType;
    property    DataSetIndex: integer read FIndex;
    property    Extent: TDSpExtent read FExtent;
    property    Driver: TDSpBaseDriver read FDriver;
  end;

  //////////////////////////////////////////////////////////////////////////////
  // PFA objects: TDSpBaseDriver, TDSpBaseBroker, TDSpSourceHandler, and TDSpPFA

  // TDSpBaseDriver is the ancestor of all the drivers available through the PFA
  TDSpBaseDriver = class
  protected
    function    Description: string; virtual; abstract;
    function    DataSets: integer; virtual; abstract;

    function    DataSetID(i: integer): integer; virtual; abstract;
    function    DataSetDescription(i: integer): string; virtual; abstract;
    function    DataSetCoordSys(i: integer): string; virtual; abstract;
    procedure   DataSetExtent(i: integer; const ext: TDSpExtent); virtual; abstract;
    function    DataSetType(i: integer): TDSpPFAFormat; virtual; abstract;
  public
    procedure   AddRasterDataSet(const r: TDSpSourceDataSet); virtual; abstract;

    procedure   GetRasterInfo(i: integer; info: PDSpRasterInfo); virtual; abstract;
    procedure   GetRasterBandInfo(i: integer; info: PDSpRasterInfo); virtual; abstract;

    procedure   ReadRasterBand(dataSet, band: integer; const into: pointer); virtual; abstract;
    procedure   ReadRasterBandRows(dataSet, band, startRow, rows: integer; const into: pointer); virtual; abstract;
    procedure   ReadRasterBandBlock(dataSet, band, startRow, startCol, rows, cols: integer; const into: pointer); virtual; abstract;
  end;

  TDSpBaseBroker = class
  protected
    FDescription: string;
    FFilter     : string;    // filter for file dialogs
    FExtension  : string;    // extension(s) supported by this driver
    FDrivers    : TList;     // of TDSpBaseDriver;
    FCaps       : TDSpPFACaps;
  public
    constructor Create;
    destructor  Destroy; override;

    function    Filter: string;
    function    SupportsExtension(ext: string): boolean; virtual;

    function    CreateSourceHandler(const fn: string): TDSpSourceHandler; virtual; abstract;
    function    CloseDriver(r: TDSpBaseDriver): boolean; virtual;

    property    Capabilities: TDSpPFACaps read FCaps;
  end;

  // A handler to a data source. The PFA issues this handler when the application
  // opens a source. The handler runs in a separate thread such that multiple
  // data sources can be accessed simultaneously by multiple handlers. Instances
  // of this class are owned by the PFA and clients will receive a reference to
  // the instance; when the handler is no longer required, the client should
  // close the reference using PFA.CloseSource, NOT delete the instance.
  TDSpSourceHandler = class(TThread)
  private
    FQueue      : TDSpThreadQueue;             // queue to receive messages in
    FLock       : TMultiReadExclusiveWriteSynchronizer; // protect access to local members
    FStatus     : TDSpStates;                  // status of the reader
  protected
    FConnect    : string;                      // connection string of the source, typically a file name
    FDescription: string;                      // general description of the source
    FRefs       : integer;                     // number of references to this source
    FDriver     : TDSpBaseDriver;              // the driver accessing the file
    FDataSets   : array of TDSpSourceDataSet;  // the data sources themselves

    procedure   Execute; override;

    procedure   OpenSource;
    function    CreateDataSet(ds: PDSpDataSetInfo; p: pointer): TDSpSourceDataSet;
    function    CloneDataSetFrom(ds: TDSpSourceDataSet): TDSpSourceDataSet;

    procedure   RenderRasterMap(msg: TDSpThreadMessage);
    procedure   ReadRasterBandDataByCoords(msg: TDSpThreadMessage);

    function    GetSourceString: string;
    function    GetNumDataSets: integer;
    procedure   AddRef;
    function    TotalRefs: integer;
  public
    constructor Create(r: TDSpBaseDriver; const source: string);
    destructor  Destroy; override;
    procedure   Close;

    procedure   QueueMessage(msg: TDSpThreadMessage);
    function    ReadLock(ndx: integer): TDSpSourceDataSet;
    function    WriteLock(ndx: integer): TDSpSourceDataSet;
    property    Status: TDSpStates read FStatus;
    property    Source: string read GetSourceString;
    property    Driver: TDSpBaseDriver read FDriver;
    property    DataSets: integer read GetNumDataSets;

    procedure   AddDataSetRef(ndx: integer);
    procedure   CloseDataSetRef(ndx: integer);
  end;

  // The data broker of the PFA. Do not instantiate this class, but access it
  // through the PFA global variable.
  TDSpPFA = class
  private
    FStaticBrokers   : array of TDSpBaseBroker;
//    FDynamicBrokers : array of TDSpBaseDriver;
    FSources         : array of TDSpSourceHandler;
    FLock            : TMultiReadExclusiveWriteSynchronizer; // protect access to local members

    function    IdentifyBroker(const ext: string): TDSpBaseBroker;
    procedure   InsertSource(s: TDSpSourceHandler);
  protected
    constructor Create;
    destructor  Destroy; override;
  public
    procedure   RegisterBroker(d: TDSpBaseBroker);
    function    FilterString(cap: TDSpPFACap): string;

    function    Open(const source: string): TDSpSourceHandler;
    function    CreateTempRaster(ds: PDSpDatasetInfo; r: PDSpRasterInfo): TDSpSourceHandler; overload;
    function    CreateTempRaster(r: TDSpSourceDataSet): TDSpSourceHandler; overload;
//    procedure   CloseSource(src: TDSpSourceHandler);
    procedure   IdleProcessing;
  end;

var
  PFA : TDSpPFA;

implementation
uses DSpRasterSource, DSpVectorSource, DSpTempRaster;

////////////////////////////////////////////////////////////////////////////////
// TDSpSourceDataSet
constructor TDSpSourceDataSet.Create(const driver: TDSpBaseDriver; dataSet: integer);
begin
  inherited Create;
  FId := 0;
  FRefs := 0;
  FLock := TMultiReadExclusiveWriteSynchronizer.Create;
  FDriver := driver;
  FIndex := dataSet;
  // Other parameters (see Open) must be set before the data set is fully defined.
end;

constructor TDSpSourceDataSet.Open(const driver: TDSpBaseDriver; dataSet: integer);
begin
  inherited Create;
  FRefs := 0;
  FLock := TMultiReadExclusiveWriteSynchronizer.Create;
  FDriver := driver;
  FIndex := dataSet;

  FID := FDriver.DataSetID(FIndex);
  FName := FDriver.DataSetDescription(FIndex);
  FCoordSys := CreateCoordinateSystemFromWKT(FDriver.DataSetCoordSys(FIndex));
  FExtent := TDSpExtent.Create(0, 0, 0, 0);
  FDriver.DataSetExtent(FIndex, FExtent);
end;

destructor TDSpSourceDataSet.Destroy;
begin
  if Assigned(FCoordSys) then FCoordSys.Release; // never free a TDSpCoordinateSystem: reference counted
  if Assigned(FExtent) then FExtent.Free;
  FLock.Free;
  inherited;
end;

procedure TDSpSourceDataSet.AddRef;
begin
  FLock.BeginWrite;
  Inc(FRefs);
  FLock.EndWrite;
end;

function TDSpSourceDataSet.CloseRef: integer;
begin
  FLock.BeginWrite;
  if FRefs > 0 then Dec(FRefs);
  if FRefs = 0 then PurgeData;
  Result := FRefs;
  FLock.EndWrite;
end;

procedure TDSpSourceDataSet.ReadLock;
begin
  FLock.BeginRead;
end;

procedure TDSpSourceDataSet.ReadUnlock;
begin
  FLock.EndRead;
end;

procedure TDSpSourceDataSet.WriteLock;
begin
  FLock.BeginWrite;
  StartWriting;
end;

procedure TDSpSourceDataSet.WriteUnlock;
begin
  EndWriting;
  FLock.EndWrite;
end;

procedure TDSpSourceDataSet.StartWriting;
begin
end;

procedure TDSpSourceDataSet.EndWriting;
begin
end;

procedure TDSpSourceDataSet.SetCoordinateSystem(cs: TDSpCoordinateSystem);
begin
  if Assigned(FCoordSys) then FCoordSys.Release;
  FCoordSys := cs;
end;

procedure TDSpSourceDataSet.SetCoordinateSystem(const wkt: string);
begin
  if Assigned(FCoordSys) then FCoordSys.Release;
  FCoordSys := CreateCoordinateSystemFromWKT(wkt);
end;

function TDSpSourceDataSet.DescribeCS(descr: TDSpDescribe): string;
begin
  if Assigned(FCoordSys) then Result := FCoordSys.Describe(descr)
  else Result := '(No coordinate system defined.)';
end;

function TDSpSourceDataSet.CoordinateSystem: TDSpCoordinateSystem;
begin
  Result := FCoordSys;
end;

procedure TDSpSourceDataSet.SetExtent(ext: TDSpExtent);
begin
  FExtent := ext.Clone;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpSourceHandler
constructor TDSpSourceHandler.Create(r: TDSpBaseDriver; const source: string);
begin
  inherited Create(true);
  FLock := TMultiReadExclusiveWriteSynchronizer.Create;
  FLock.BeginWrite;
  try
    FRefs := 1;
    FDriver := r;
    FStatus := [dspOk, dspReading, dspBusy];
    FQueue := TDSpThreadQueue.Create;
  finally
    FLock.EndWrite;
  end;
end;

destructor TDSpSourceHandler.Destroy;
var i: integer;
begin
  FLock.BeginWrite;
  try
    if Length(FDataSets) > 0 then
      for i := 0 to High(FDataSets) do
        if Assigned(FDataSets[i]) then FDataSets[i].Free;
    Finalize(FDataSets);
    FDriver.Free;
    FQueue.Free;
  finally
    FLock.EndWrite;
  end;
  inherited;
end;

procedure TDSpSourceHandler.Close;
begin
  FLock.BeginWrite;
  try
    if FRefs > 0 then Dec(FRefs);
  finally
    FLock.EndWrite;
  end;
end;

procedure TDSpSourceHandler.Execute;
var msg: TDSpThreadMessage;
begin
  while not Terminated do begin
    msg := TDSpThreadMessage(FQueue.Get);

    if msg = nil then begin
      FStatus := [dspOk, dspIdle];  // no lock required here as FStatus is private
      Suspend; 
      FStatus := [dspOk, dspBusy];
      Continue;                     // after waking up, start by getting a new message
    end;

    try
      case msg^.Msg of
        DSPMSG_NULL             : ; // dummy message, just set the event
        DSPMSG_TERMINATE        : Terminate;
        DSPMSG_SOURCEOPEN       : OpenSource;
        DSPMSG_RENDERRASTERMAP  : RenderRasterMap(msg);
        DSPMSG_BANDREADCOORDS   : ReadRasterBandDataByCoords(msg);
      else Assert(false, 'Unhandled message id in TDSpSourceHandler.Execute');
      end;
    except
      msg^.Msg := DSPMSG_ERROR;
      FStatus := [dspError];
    end;

    msg^.Event.SetEvent;
  end;
end;

procedure TDSpSourceHandler.OpenSource;
var i, num: integer;
    dst: TDSpPFAFormat;
begin
  FStatus := FStatus + [dspReading];
  FLock.BeginWrite;
  try
    // Get global data source info
    FDescription := FDriver.Description;
    num := FDriver.DataSets;
    SetLength(FDataSets, num);
    if num = 0 then Exit;

    // Create the data sets
    for i := 0 to num - 1 do begin
      dst := FDriver.DataSetType(i);
      case dst of
        dspPFARaster: FDataSets[i] := TDSpRasterSource.Open(FDriver, i);
        dspPFAVector: ;
      end;
    end;
  finally
    FLock.EndWrite;
    FStatus := [dspOk, dspIdle];
  end;
end;

function TDSpSourceHandler.CreateDataSet(ds: PDSpDataSetInfo; p: pointer): TDSpSourceDataSet;
begin
  Assert(false, 'Not yet implemented');
  Result := nil;
end;

// CloneDataSetFrom is used to create a data set that is compatible with an
// existing data set: same extent, etc. For rasters this means same rows,
// columns, and pixel size. A source handler can handle multiple data sets by
// subsequent calls to this method.
function TDSpSourceHandler.CloneDataSetFrom(ds: TDSpSourceDataSet): TDSpSourceDataSet;
var ndx: integer;
begin
  FStatus := [dspBusy];
  FLock.BeginWrite;
  try
    ndx := Length(FDataSets);
    SetLength(FDataSets, ndx+1);
    Result := ds.Clone(FDriver, ndx);
    FDataSets[ndx] := Result;
  finally
    FLock.EndWrite;
  end;
  FStatus := [dspIdle];
end;

procedure TDSpSourceHandler.RenderRasterMap(msg: TDSpThreadMessage);
begin
  Assert(FDataSets[msg^.DataSourceIndex] is TDSpRasterSource);
  TDSpRasterSource(FDataSets[msg^.DataSourceIndex]).RenderBand(msg^.Parameter);
end;

procedure TDSpSourceHandler.ReadRasterBandDataByCoords(msg: TDSpThreadMessage);
begin
  Assert(FDataSets[msg^.DataSourceIndex] is TDSpRasterSource);
  TDSpRasterSource(FDataSets[msg^.DataSourceIndex]).ReadBandByCoords(msg^.Parameter);
end;

procedure TDSpSourceHandler.QueueMessage(msg: TDSpThreadMessage);
begin
  FQueue.Put(msg);
  Resume;
end;

function TDSpSourceHandler.ReadLock(ndx: integer): TDSpSourceDataSet;
begin
  Assert(ndx in [0..High(FDataSets)]);
  Result := FDataSets[ndx];
  Result.ReadLock;
end;

function TDSpSourceHandler.WriteLock(ndx: integer): TDSpSourceDataSet;
begin
  Assert(ndx in [0..High(FDataSets)]);
  Result := FDataSets[ndx];
  Result.WriteLock;
end;

function TDSpSourceHandler.GetSourceString: string;
begin
  Result := FConnect;
end;

function TDSpSourceHandler.GetNumDataSets: integer;
begin
  Result := Length(FDataSets);
end;

procedure TDSpSourceHandler.AddRef;
begin
  Inc(FRefs);
end;

procedure TDSpSourceHandler.AddDataSetRef(ndx: integer);
begin
  Assert(ndx in [0..High(FDataSets)]);
  FDataSets[ndx].AddRef;
end;

procedure TDSpSourceHandler.CloseDataSetRef(ndx: integer);
begin
  Assert(ndx in [0..High(FDataSets)]);
  FDataSets[ndx].CloseRef;
end;

function TDSpSourceHandler.TotalRefs: integer;
var i: integer;
    ds: TDSpSourceDataSet;
begin
  FLock.BeginRead;
  try
    Result := FRefs;
    if Length(FDataSets) > 0 then
      for i := 0 to High(FDataSets) do begin
        ds := ReadLock(i);
        Inc(Result, ds.FRefs);
        ds.ReadUnlock;
      end;
  finally
    FLock.EndRead;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpBaseBroker - ancestor of all PFA drivers
constructor TDSpBaseBroker.Create;
begin
  // descendants should set the FDescription and FExtension members in their constructors
  inherited;
  FDrivers := TList.Create;
end;

destructor TDSpBaseBroker.Destroy;
begin
  // NOTE: descendants must free any open readers before invoking this destructor
  FDrivers.Clear;
  FDrivers.Free;
  inherited;
end;

function TDSpBaseBroker.Filter: string;
begin
  Result := FDescription + '|' + FFilter;
end;

function TDSpBaseBroker.SupportsExtension(ext: string): boolean;
begin
  Result := FExtension = ext;
end;

// Basic termination of the driver thread. Override this virtual method for
// broker specific action, but ALWAYS call this method from the descendant
// method.
function TDSpBaseBroker.CloseDriver(r: TDSpBaseDriver): boolean;
begin
  Result := true;
  Free;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpPFA
constructor TDSpPFA.Create;
begin
  inherited;
  FLock := TMultiReadExclusiveWriteSynchronizer.Create;
  SetLength(FStaticBrokers, 1);
  FStaticBrokers[0] := TDSpTempRasterBroker.Create;
end;

destructor TDSpPFA.Destroy;
var i: integer;
begin
  FLock.BeginWrite;
  try
    if Length(FSources) > 0 then
      for i := 0 to High(FSources) do FSources[i].Free;
    Finalize(FSources);

    for i := High(FStaticBrokers) downto 0 do FStaticBrokers[i].Free;
    Finalize(FStaticBrokers);
  finally
    FLock.EndWrite;
  end;
  FLock.Free;
  inherited;
end;

// Register a static broker with the PFA. Static brokers cannot be removed from
// the PFA (hence the name). Use dynamic brokers in DLLs for resource
// conservation (mostly useful for less used data formats).
procedure TDSpPFA.RegisterBroker(d: TDSpBaseBroker);
begin
  // Lock is not required here because this function will only be called upon
  // initialization.
  SetLength(FStaticBrokers, Length(FStaticBrokers)+1);
  FStaticBrokers[High(FStaticBrokers)] := d;
end;

function TDSpPFA.FilterString(cap: TDSpPFACap): string;
var i: integer;
begin
  FLock.BeginRead;
  try
    Result := '';
    for i := 0 to High(FStaticBrokers) do
      if cap in FStaticBrokers[i].Capabilities then begin
        if Result <> '' then Result := Result + '|';
        Result := Result + FStaticBrokers[i].Filter;
      end;
  finally
    FLock.EndRead;
  end;
end;

procedure TDSpPFA.InsertSource(s: TDSpSourceHandler);
begin
  FLock.BeginWrite;
  try
    SetLength(FSources, Length(FSources) + 1);
    FSources[High(FSources)] := s;
  finally
    FLock.EndWrite;
  end;
end;

// Create a handler to a data source
function TDSpPFA.Open(const source: string): TDSpSourceHandler;
var broker: TDSpBaseBroker;
    i: integer;
begin
  FLock.BeginRead;
  Result := nil;
  try
    // First check if the source is already handled. If found, return that
    // handler and increase its ref count.
    if Length(FSources) > 0 then
      for i := 0 to High(FSources) do
        if source = FSources[i].Source then begin
          Result := FSources[i];
          Result.AddRef;
          Exit;
        end;

    // Find the right broker
    broker := IdentifyBroker(Copy(ExtractFileExt(source), 2, 1000));
    if broker = nil then
      raise Exception.Create('Format not supported by the PFA');

    // Create the handler and register it with the PFA
    Result := broker.CreateSourceHandler(source);
    Result.OpenSource;
  finally
    FLock.EndRead;
  end;
  Assert(Result <> nil);
  InsertSource(Result);
end;

// Create a temporary raster
function TDSpPFA.CreateTempRaster(ds: PDSpDatasetInfo; r: PDSpRasterInfo): TDSpSourceHandler;
begin
  Result := FStaticBrokers[0].CreateSourceHandler('');
  Result.CreateDataSet(ds, r);
  InsertSource(Result);
end;

function TDSpPFA.CreateTempRaster(r: TDSpSourceDataSet): TDSpSourceHandler;
begin
  Result := FStaticBrokers[0].CreateSourceHandler('');
  Result.CloneDataSetFrom(r);
  InsertSource(Result);
end;

function TDSpPFA.IdentifyBroker(const ext: string): TDSpBaseBroker;
var e: string;
    i: integer;
begin
  // This method has to change to allow for source strings that do not have an
  // extension, such as database connection strings
  FLock.BeginRead;
  try
    e := UpperCase(ext);
    for i := 0 to High(FStaticBrokers) do
      if FStaticBrokers[i].SupportsExtension(e) then begin
        Result := FStaticBrokers[i];
        Exit;
      end;
    Result := nil;
  finally
    FLock.EndRead;
  end;
end;

// This method is called from TDSpatial.IdleProcessing when the application is
// idle. The list of sources is scanned to find any sources to which all
// references have been dropped. These sources are then deleted and the list
// updated.
procedure TDSpPFA.IdleProcessing;
var i, j, max, del: integer;
begin
  if Length(FSources) = 0 then Exit;

  FLock.BeginWrite;
  try
    max := High(FSources);
    del := 0;
    for i := 0 to max do  // First delete any unused sources...
      if FSources[i].TotalRefs = 0 then begin
        FSources[i].Free;
        FSources[i] := nil;
        Inc(del);
      end;
    if del > 0 then begin // ... then update the list if needed
      if max > 0 then begin
        j := max;
        while FSources[j] = nil do Dec(j); // j points to the highest index source
        for i := 0 to max - del do
          if FSources[i] = nil then begin
            FSources[i] := FSources[j];
            repeat Dec(j) until FSources[j] <> nil;
          end;
      end;
      SetLength(FSources, max - del + 1);
    end;
  finally
    FLock.EndWrite;
  end;
end;

initialization
  PFA := TDSpPFA.Create;
finalization
  if Assigned(PFA) then PFA.Free;
end.
