{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRasterData.pas. The Original Code is available       }
{ through the home page of DSpatial at http://dspatial.sourceforge.net.        }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements the structures that hold raster data in memory. Storage }
{ is organized by the size of the data type of the raster band; i.e. dspInt32  }
{ and dspFloat32 data types share the same structure for 32-bit data. More     }
{ specialized classes that implement features like compression, quad-trees, or }
{ tiling, descend from these bit-size ancestors, and they may not be available }
{ for all data types.                                                          }
{                                                                              }
{ These storage classes are "dumb" in the sense that they do not "know" about  }
{ the data other than its size.                                                }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 16                                             }
{                                                                              }
{ Dependencies  : DSpPFA                                                       }
{                 DSpGlobals                                                   }
{                 Math (Ceil)                                                  }
{                 SysUtils (Exception)                                         }
{                 Windows (TRect)                                              }
{                                                                              }
{ Version       : 0.4  03Nov16 Added code to purge unused data. Combined Render}
{                              and RenderFullResolution. Added ReadDataByCoords}
{                              to read source data into a block of memory.     }
{                 0.3  03Nov07 PFA overhaul                                    }
{                 0.2  03Aug16 Added RenderFullResolution                      }
{                 0.1  03Aug03 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpRasterData;
{$I DSpatial.inc}
interface
uses Windows, DSpGlobals, DSpPFA;
type
  // Abstract ancestor of all raster storage classes
  TDSpRasterData = class
  protected
    FRaster   : TObject;
    FBand     : integer;
    FDriver   : TDSpBaseDriver;
    FDataSize : integer;
  public
    constructor Create(raster: TObject; band: integer);

    procedure   SetData(p: pointer); virtual; abstract;
    procedure   PurgeData; virtual; abstract;
    function    GetAllData: pointer; virtual; abstract;
    function    GetRowData(const r: integer): pointer; virtual; abstract;
    function    ValueAsFloat64(row: pointer; c: integer): TDSpFloat64; virtual; abstract;

    procedure   Render(params: PDSpRenderRasterMap);
    procedure   ReadDataByCoords(params: PDSpMsgReadRasterBandCoords); virtual; abstract;
  end;

  TDSpRasterDataFlat = class(TDSpRasterData)
  protected
    FData     : pointer;
  public
    destructor  Destroy; override;

    procedure   SetData(p: pointer); override;
    procedure   PurgeData; override;
    function    GetAllData: pointer; override;
    function    GetRowData(const r: integer): pointer; override;
    procedure   ReadDataByCoords(params: PDSpMsgReadRasterBandCoords); override;
  end;

  TDSpRasterData8Flat = class(TDSpRasterDataFlat)
  public
    constructor Create(raster: TObject; band: integer);
  end;

  TDSpRasterInt8Flat = class(TDSpRasterData8Flat)
  public
    function    ValueAsFloat64(row: pointer; c: integer): TDSpFloat64; override;
  end;

  TDSpRasterUInt8Flat = class(TDSpRasterData8Flat)
  public
    function    ValueAsFloat64(row: pointer; c: integer): TDSpFloat64; override;
  end;

  TDSpRasterData16Flat = class(TDSpRasterDataFlat)
  public
    constructor Create(raster: TObject; band: integer);
  end;

  TDSpRasterInt16Flat = class(TDSpRasterData16Flat)
  public
    function    ValueAsFloat64(row: pointer; c: integer): TDSpFloat64; override;
  end;

  TDSpRasterUInt16Flat = class(TDSpRasterData16Flat)
  public
    function    ValueAsFloat64(row: pointer; c: integer): TDSpFloat64; override;
  end;

  TDSpRasterData32Flat = class(TDSpRasterDataFlat)
  public
    constructor Create(raster: TObject; band: integer);
  end;

  TDSpRasterInt32Flat = class(TDSpRasterData32Flat)
  public
    function    ValueAsFloat64(row: pointer; c: integer): TDSpFloat64; override;
  end;

  TDSpRasterUInt32Flat = class(TDSpRasterData32Flat)
  public
    function    ValueAsFloat64(row: pointer; c: integer): TDSpFloat64; override;
  end;

  TDSpRasterFloat32Flat = class(TDSpRasterData32Flat)
  public
    function    ValueAsFloat64(row: pointer; c: integer): TDSpFloat64; override;
  end;

  TDSpRasterFloat64Flat = class(TDSpRasterDataFlat)
  public
    constructor Create(raster: TObject; band: integer);
    function    ValueAsFloat64(row: pointer; c: integer): TDSpFloat64; override;
  end;

  function CreateStorage(raster: TObject; band: integer; dt: TDSpDataType): TDSpRasterData;

implementation
uses SysUtils, Math, DSpUtils, DSpRasterSource;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterData
constructor TDSpRasterData.Create(raster: TObject; band: integer);
begin
  inherited Create;
  FRaster := raster;
  FBand := band;
  FDriver := TDSpRasterSource(FRaster).Driver;
end;

procedure TDSpRasterData.Render(params: PDSpRenderRasterMap);
var raster    : TDSpRasterSource;
    map       : TDSpRenderMap;
    overlap   : TDSpExtent;
    miss, val, scale, min, max, incr: TDSpFloat64;
    rows, cols: integer;
    i, j      : integer;
    r, c      : integer;
    startCol, endCol: integer;
    row       : pointer;
    run, v, h : TDSpFloat64;
    rendVal   : ^word; //TDSpWordArray;
begin
  raster := FRaster as TDSpRasterSource;

  map := params^.Map;
  if FloatsSame(raster.MissingValue[FBand], 0) then miss := 0
  else miss := 1 / raster.MissingValue[FBand];
  rows := raster.Rows;
  cols := raster.Columns;
  min := params^.Min;
  max := params^.Max;
  scale := 65535 / (max - min);

  if params^.Pixel < 0 then begin
    // Render at the full resolution of the raster band
    Assert(map.Rows = rows);
    Assert(map.Columns = cols);

    for r := 0 to rows - 1 do begin
      row := GetRowData(r);
      rendVal := @map.Row[r]^;
      for c := 0 to cols - 1 do begin
        val := ValueAsFloat64(row, c);
        if not FloatsSameInv(val, miss) then begin
          if val <= min then rendVal^ := 1
          else if val >= max then rendVal^ := 65535
          else rendVal^ := Ceil((val - min) * scale);
        end;
        Inc(rendVal);
      end;
    end;
  end else begin
    // Render according to the viewer geometry
    overlap := raster.Extent.Intersection(params^.Area);
    Assert(overlap <> nil);
    incr := 1 / raster.PixelSize; // temporary
    v := (raster.Extent.North - overlap.North) * incr;
    h := (overlap.West - raster.Extent.West) * incr;
    overlap.Free;
    incr := params^.Pixel * incr; // increment of source pixels for each map pixel
    startCol := params^.MapArea.Left;
    endCol := params^.MapArea.Right;

    for i := params^.MapArea.Top to params^.MapArea.Bottom do begin
      r := Round(v);
      if r < 0 then r := 0;
      if r >= rows then r := rows - 1;
      row := GetRowData(r);
      run := h;
      rendVal := pointer(integer(map.Row[i]) + startCol * 2);
      for j := startCol to endCol do begin
        c := Round(run);
        if c < 0 then c := 0;
        if c >= cols then c := cols - 1;
        val := ValueAsFloat64(row, c);
        if not FloatsSameInv(val, miss) then begin
          if val <= min then rendVal^ := 1
          else if val >= max then rendVal^ := 65535
          else rendVal^ := Ceil((val - min) * scale);
        end;
        run := run + incr;
        Inc(rendVal);
      end;
      v := v + incr;
    end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterDataFlat
destructor TDSpRasterDataFlat.Destroy;
begin
  if FData <> nil then FreeMem(FData);
end;

procedure TDSpRasterDataFlat.SetData(p: pointer);
begin
  if FData <> nil then Freemem(FData);
  FData := p;
end;

procedure TDSpRasterDataFlat.PurgeData;
begin
  if FData <> nil then begin
    FreeMem(FData);
    FData := nil;
  end;
end;

function TDSpRasterDataFlat.GetAllData: pointer;
var raster: TDSpRasterSource;
begin
  // Check if data is available. If not, go get it.
  if FData = nil then begin
    raster := FRaster as TDSpRasterSource;
    FData := AllocMem(FDataSize * raster.Rows * raster.Columns);
    FDriver.ReadRasterBand(raster.DataSetIndex, FBand, FData);
  end;
  Result := FData;
end;

function TDSpRasterDataFlat.GetRowData(const r: integer): pointer;
begin
  Assert((r >= 0) and (r < TDSpRasterSource(FRaster).Rows));
  GetAllData;  // make sure the data is loaded
  Result := pointer(integer(FData) + r * TDSpRasterSource(FRaster).Columns * FDataSize);
end;

procedure TDSpRasterDataFlat.ReadDataByCoords(params: PDSpMsgReadRasterBandCoords);
var i, row, endrow, columns, pixels: integer;
    currSource, currData: pointer;
begin
  GetAllData;  // make sure the data is loaded

  row := params^.Coords.Top;
  endrow := params^.Coords.Bottom;
  columns := TDSpRasterSource(FRaster).Columns;

  if (params^.Coords.Left = 0) and (params^.Coords.Right = columns - 1) then
    // Read a number of complete rows
    Move(pointer(integer(FData) + row * columns * FDataSize)^, params^.Data^, (endrow - row + 1) * columns * FDataSize)
  else begin
    // Read a block of data
    currSource := pointer(integer(FData) + (row * columns + params^.Coords.Left) * FDataSize);
    currData := params^.Data;
    pixels := params^.Coords.Right - params^.Coords.Left + 1; // number of pixels being read
    for i := row to endrow do begin
      Move(currSource^, currData^, pixels * FDataSize);
      Inc(integer(currSource), columns * FDataSize);
      Inc(integer(currData), pixels * FDataSize);
    end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterData8Flat
constructor TDSpRasterData8Flat.Create(raster: TObject; band: integer);
begin
  inherited;
  FDataSize := 1;
end;

function TDSpRasterInt8Flat.ValueAsFloat64(row: pointer; c: integer): TDSpFloat64;
begin
  Result := TDSpInt8(pointer(integer(row) + c)^);
end;

function TDSpRasterUInt8Flat.ValueAsFloat64(row: pointer; c: integer): TDSpFloat64;
begin
  Result := TDSpUInt8(pointer(integer(row) + c)^);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterData16Flat
constructor TDSpRasterData16Flat.Create(raster: TObject; band: integer);
begin
  inherited;
  FDataSize := 2;
end;

function TDSpRasterInt16Flat.ValueAsFloat64(row: pointer; c: integer): TDSpFloat64;
begin
  Result := TDSpInt16(pointer(integer(row) + c * 2)^);
end;

function TDSpRasterUInt16Flat.ValueAsFloat64(row: pointer; c: integer): TDSpFloat64;
begin
  Result := TDSpUInt16(pointer(integer(row) + c * 2)^);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterData32Flat
constructor TDSpRasterData32Flat.Create(raster: TObject; band: integer);
begin
  inherited;
  FDataSize := 4;
end;

function TDSpRasterInt32Flat.ValueAsFloat64(row: pointer; c: integer): TDSpFloat64;
begin
  Result := TDSpInt32(pointer(integer(row) + c * 4)^);
end;

function TDSpRasterUInt32Flat.ValueAsFloat64(row: pointer; c: integer): TDSpFloat64;
begin
  Result := TDSpUInt32(pointer(integer(row) + c * 4)^);
end;

function TDSpRasterFloat32Flat.ValueAsFloat64(row: pointer; c: integer): TDSpFloat64;
begin
  Result := TDSpFloat32(pointer(integer(row) + c * 4)^);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterFloat64Flat
constructor TDSpRasterFloat64Flat.Create(raster: TObject; band: integer);
begin
  inherited;
  FDataSize := 8;
end;

function TDSpRasterFloat64Flat.ValueAsFloat64(row: pointer; c: integer): TDSpFloat64;
begin
  Result := TDSpFloat64(pointer(integer(row) + c * 8)^);
end;

////////////////////////////////////////////////////////////////////////////////
// Global function to create the right storage class for the data
function CreateStorage(raster: TObject; band: integer; dt: TDSpDataType): TDSpRasterData;
begin
  case dt of
    dspInt8   : Result := TDSpRasterInt8Flat.Create(raster, band);
    dspUInt8  : Result := TDSpRasterUInt8Flat.Create(raster, band);
    dspInt16  : Result := TDSpRasterInt16Flat.Create(raster, band);
    dspUInt16 : Result := TDSpRasterUInt16Flat.Create(raster, band);
    dspInt32  : Result := TDSpRasterInt32Flat.Create(raster, band);
    dspUInt32 : Result := TDSpRasterUInt32Flat.Create(raster, band);
    dspFloat32: Result := TDSpRasterFloat32Flat.Create(raster, band);
    dspFloat64: Result := TDSpRasterFloat64Flat.Create(raster, band);
  else Result := nil;
  end;
end;

end.
