{******************************************************************************}
{                                                                              }
{ Project DSpatial - Geospatial analysis for Delphi                            }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/.}
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is DSpRenderer.pas. The Original Code is available through }
{ the home page of DSpatial at http://dspatial.sourceforge.net.                }
{                                                                              }
{ The Initial Developer of the Original Code is Patrick Van Laake              }
{ (pvanlaake@users.sourceforge.net). Portions created by the Initial Developer }
{ are Copyright (C) of the Initial Developer. Changes to the Original Code     }
{ available through DSpatial are Copyright (C) of the Contributing Authors.    }
{                                                                              }
{******************************************************************************}
{                                                                              }
{ This unit implements the different kinds of renderers for data sets inside a }
{ DSpViewer. Depending on the characteristics of the data set a particular     }
{ renderer is used. Renderers can be organized into a DSpLegend component for  }
{ display next to a DSpViewer. The raster renderer makes use of a helper class }
{ which defines the actual transformation from raster band values to colours.  }
{ These so-called raster renderer transforms (RRT) are located in the unit     }
{ DSpRRT.                                                                      }
{                                                                              }
{ Unit owner    : Patrick Van Laake                                            }
{ Last modified : 2003 November 14                                             }
{                                                                              }
{ Dependencies  : DSpRaster                                                    }
{                 DSpVector                                                    }
{                 DSpRRT                                                       }
{                 DSpViewer                                                    }
{                 DSpPFA (TDSpThreadMessage)                                   }
{                 DSpGraphics (SaveAsImage)                                    }
{                 DSpUtils (FloatsSame)                                        }
{                 DSpGlobals                                                   }
{                 Graphics (TBitmap)                                           }
{                 SyncObjs (TEvent)                                            }
{                 Classes (Rect function)                                      }
{                 Windows (TRect, IntersectRect)                               }
{                 Math (Ceil)                                                  }
{                                                                              }
{ Version       : 0.5  03Nov14 Min/Max value of raster bands supported.        }
{                              Rendering of raster bands done locally.         }
{                 0.4  03Nov07 PFA overhaul                                    }
{                 0.3  03Sep25 Added SaveToImage                               }
{                 0.2  03Aug14 Rendering engine overhauled, many changes.      }
{                 0.1  03Jul18 Initial coding.                                 }
{                                                                              }
{******************************************************************************}
unit DSpRenderer;
{$I DSpatial.inc}
interface
uses Windows, Graphics, SyncObjs, DSpGlobals, DSpPFA, DSpRaster, DSpVector, DSpRRT;
type
  TDSpRenderer = class
  protected
    FLabel    : string;
    FViewer   : TObject;       // = TDSpViewer
    FTotalArea: TDSpExtent;    // total extent of bands in renderer
    FAOI      : TDSpExtent;    // area-of-interest
    FViewArea : TRect;         // area of the renderer in pixel space

    function    GetExtent: TDSpExtent;
    procedure   SetExtent(ext: TDSpExtent);
  public
    constructor Create(const viewer: TObject; lab: string);
    destructor  Destroy; override;

    procedure   DissociateFromViewer;
    procedure   ResetArea; virtual; abstract;
    procedure   SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64); virtual;
    procedure   Paint; virtual; abstract;
    procedure   SaveImage(const fn: string); virtual; abstract;
    property    DisplayArea: TDSpExtent read GetExtent write SetExtent;
    property    Name: string read FLabel write FLabel;
  end;

  TDSpRasterRenderer = class(TDSpRenderer)
  protected
    FBitmap   : TBitmap;
    FRenderMap: TDSpRenderMap;

    procedure   ViewAreaToRendererArea(view: TRect; var rend: TRect);
    procedure   DoRender(const clip: TRect); virtual; abstract;
  public
    constructor Create(const viewer: TObject; lab: string);
    destructor  Destroy; override;

    procedure   SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64); override;
    procedure   Paint; override;
    procedure   DeleteRasterBand(b: TDSpRasterBand); virtual; abstract;
  end;

  TDSpBandRenderer = record
                       Band  : TDSpRasterBand;
                       Msg   : TDSpThreadMessage;
                       Status: TDSpState;
                     end;

  TDSpRasterRendererSimple = class(TDSpRasterRenderer)
  // A single raster band determines the entire RGB-triplet
  protected
    FData       : array of TDSpBandRenderer;
    FRenderMin,
    FRenderMax  : TDSpFloat64;
    FTrans      : TDSpRRTSingleBand;

    procedure   SetTrans(t: TDSpRRTSingleBand);
    function    GetTrans: TDSpRRTSingleBand;
    procedure   SetMin(val: TDSpFloat64);
    function    GetMin: TDSpFloat64;
    procedure   SetMax(val: TDSpFloat64);
    function    GetMax: TDSpFloat64;
    function    GetNumBands: integer;
    function    GetBand(b: integer): TDSpRasterBand;

    procedure   WaitFor(ndx: integer);
  public
    constructor Create(const viewer: TObject; lab: string);
    destructor  Destroy; override;

    procedure   AddRasterBand(band: TDSpRasterBand);
    procedure   DeleteRasterBand(band: TDSpRasterBand); override;
    procedure   DoRender(const clip: TRect); override;
    procedure   SaveImage(const fn: string); override;
    procedure   SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64); override;
    procedure   ResetArea; override;
    procedure   Range(min, max: TDSpFloat64);

    property    Transformation: TDSpRRTSingleBand read GetTrans write SetTrans;
    property    Min: TDSpFloat64 read GetMin write SetMin;
    property    Max: TDSpFloat64 read GetMax write SetMax;
    property    Bands: integer read GetNumBands;
    property    Band[b: integer]: TDSpRasterBand read GetBand;
  end;

  TDSpFeatureRenderer = class(TDSpRenderer)
  protected
    FDataSet : TDSpVector;

  public
    constructor Create(const viewer: TObject; lab: string);

    procedure   Paint; override;
    procedure   SaveImage(const fn: string); override;
  end;

implementation
uses Math, Classes, DSpViewer, DSpUtils, DSpGraphics;

constructor TDSpRenderer.Create(const viewer: TObject; lab: string);
begin
  Assert(viewer is TDSpViewer);
  FViewer := viewer;
  FLabel := lab;
end;

destructor TDSpRenderer.Destroy;
begin
  if Assigned(FTotalArea) then FTotalArea.Free;
  if Assigned(FAOI) then FAOI.Free;
  inherited;
end;

procedure TDSpRenderer.DissociateFromViewer;
begin
  (FViewer as TDSpViewer).DeleteRenderer(self);
end;

procedure TDSpRenderer.SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64);
var ext: TDSpExtent;
begin
  if not Assigned(FTotalArea) then Exit;
  if Assigned(FAOI) then ext := FAOI else ext := FTotalArea;
  FViewArea := ext.PixelExtent(view, area, pixel);
end;

function TDSpRenderer.GetExtent: TDSpExtent;
begin
  Result := nil;
  if Assigned(FAOI) then Result := FAOI
  else if Assigned(FTotalArea) then Result := FTotalArea
  else Assert(false, 'DSpRenderer.GetExtent: Not initialized');
end;

procedure TDSpRenderer.SetExtent(ext: TDSpExtent);
begin
  if Assigned(FAOI) then FAOI.Free;
  FAOI := ext.Clone;
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterRenderer
constructor TDSpRasterRenderer.Create(const viewer: TObject; lab: string);
begin
  inherited;
  FBitmap := TBitmap.Create;
  FBitmap.PixelFormat := pf24bit;
  FBitmap.Transparent := true;
  FRenderMap := TDSpRenderMap.Create(0, 0);
end;

destructor TDSpRasterRenderer.Destroy;
begin
  FBitmap.Free;
  FRenderMap.Free;
  inherited;
end;

procedure TDSpRasterRenderer.SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64);
begin
  inherited;
  if FViewArea.Left > FViewArea.Right then FRenderMap.Resize(0, 0)
  else FRenderMap.Resize(FViewArea.Bottom - FViewArea.Top + 1, FViewArea.Right - FViewArea.Left + 1);
end;

procedure TDSpRasterRenderer.ViewAreaToRendererArea(view: TRect; var rend: TRect);
begin
  rend.Top := view.Top - FViewArea.Top;
  rend.Bottom := view.Bottom - FViewarea.Top;
  rend.Left := view.Left - FViewArea.Left;
  rend.Right := view.Right - FViewArea.Left;
end;

procedure TDSpRasterRenderer.Paint;
var clip: TRect;
begin
  if not Assigned(FTotalArea) then Exit;
  if not IntersectRect(clip, (FViewer as TDSpViewer).Canvas.ClipRect, FViewArea) then Exit;

  // Create the bitmap to fill the "dirty" area
  FBitmap.Width := clip.Right - clip.Left + 1;
  FBitmap.Height := clip.Bottom - clip.Top + 1;

  // Dispatch to the appropriate sub-class
  DoRender(clip);
  (FViewer as TDSpViewer).Canvas.Draw(clip.Left, clip.Top, FBitmap);
end;

////////////////////////////////////////////////////////////////////////////////
// TDSpRasterRendererSimple
constructor TDSpRasterRendererSimple.Create(const viewer: TObject; lab: string);
begin
  inherited;
  FRenderMin := 0;
  FRenderMax := 1000;
  FTrans := TDSpRRTHotToCold.Create;
end;

destructor TDSpRasterRendererSimple.Destroy;
var i: integer;
begin
  FTrans.Free;
  if Length(FData) > 0 then
    for i := 0 to High(FData) do begin
      Dispose(PDSpRenderRasterMap(FData[i].Msg^.Parameter));
      FData[i].Msg^.Event.Free;
      Dispose(FData[i].Msg);
    end;
  Finalize(FData);
  inherited;
end;

procedure TDSpRasterRendererSimple.AddRasterBand(band: TDSpRasterBand);
var min, max: TDSpFloat64;
    i: integer;
begin
  SetLength(FData, Length(FData)+1);
  i := High(FData);
  FData[i].Band := band;
  New(FData[i].Msg);
  FData[i].Msg^.DataSourceIndex := band.Raster.SourceIndex;
  FData[i].Msg^.Event := TEvent.Create(nil, false, false, '');
  FData[i].Msg^.Msg := DSPMSG_RENDERRASTERMAP;;
  New(PDSpRenderRasterMap(FData[i].Msg^.Parameter));
  PDSpRenderRasterMap(FData[i].Msg^.Parameter)^.Band := band.BandIndex;
  PDSpRenderRasterMap(FData[i].Msg^.Parameter)^.Map := FRenderMap;
  FData[i].Status := dspIdle;

  min := band.Min;
  max := band.Max;
  if Length(FData) = 1 then begin
    FTotalArea := band.Raster.Extent.Clone;
    if min < max then begin
      FRenderMin := min;
      FRenderMax := max;
    end;
  end else begin
    FTotalArea.UnionWith(band.Raster.Extent);
    if min < max then begin
      if min < FRenderMin then FRenderMin := min;
      if max > FRenderMax then FRenderMax := max;
    end;
  end;
  (FViewer as TDSpViewer).MergeExtent(FTotalArea);
end;

procedure TDSpRasterRendererSimple.DeleteRasterBand(band: TDSpRasterBand);
var i, max: integer;
begin
  if Length(FData) = 0 then Exit;
  max := High(FData);
  for i := 0 to max do
    if FData[i].Band = band then begin
      Dispose(PDSpRenderRasterMap(FData[i].Msg^.Parameter));
      FData[i].Msg^.Event.Free;
      Dispose(FData[i].Msg);
      if i < max then FData[i] := FData[max];
      SetLength(FData, max);
      Break;
    end;
  ResetArea;
end;

procedure TDSpRasterRendererSimple.SetTrans(t: TDSpRRTSingleBand);
begin
  FTrans.Free;
  FTrans := t;
end;

function TDSpRasterRendererSimple.GetTrans: TDSpRRTSingleBand;
begin
  Result := FTrans;
end;

function TDSpRasterRendererSimple.GetNumBands: integer;
begin
  Result := Length(FData);
end;

function TDSpRasterRendererSimple.GetBand(b: integer): TDSpRasterBand;
begin
  Assert((b >= 0) and (b < Length(FData)));
  Result := FData[b].Band;
end;

procedure TDSpRasterRendererSimple.SetMapArea(view: TDSpExtent; area: TRect; pixel: TDSpFloat64);
var i: integer;
    overlap : TDSpExtent;
begin
  inherited; // Set FViewArea

  for i := 0 to High(FData) do begin
    overlap := FData[i].Band.Raster.Extent.Intersection(view);
    if overlap <> nil then begin
      PDSpRenderRasterMap(FData[i].Msg^.Parameter)^.MapArea := overlap.PixelExtent(view, area, pixel);
      with PDSpRenderRasterMap(FData[i].Msg^.Parameter)^ do begin
        if MapArea.Left > MapArea.Right then Continue; // band not in view
        ViewAreaToRendererArea(MapArea, MapArea);
      end;
      PDSpRenderRasterMap(FData[i].Msg^.Parameter)^.Area := view;
      PDSpRenderRasterMap(FData[i].Msg^.Parameter)^.Pixel := pixel;
      PDSpRenderRasterMap(FData[i].Msg^.Parameter)^.Min := FRenderMin;
      PDSpRenderRasterMap(FData[i].Msg^.Parameter)^.Max := FRenderMax;
      overlap.Free;
      FData[i].Status := dspPending;
      FData[i].Band.Raster.Handler.QueueMessage(FData[i].Msg);
    end;
  end;
end;

procedure TDSpRasterRendererSimple.WaitFor(ndx: integer);
begin
  if FData[ndx].Status = dspPending then
    if FData[ndx].Msg^.Event.WaitFor($FFFFFFFF) = wrSignaled then
      FData[ndx].Status := dspIdle
    else Assert(false, 'Wait event in TDSpRasterRendererSimple unexpectedly failed');
end;

procedure TDSpRasterRendererSimple.DoRender(const clip: TRect);
var i: integer;
    area: TRect;
begin
  if Length(FData) = 0 then Exit;

  // Prep the bitmap
  FBitmap.TransparentColor := FTrans.Transparent;
  FBitMap.Canvas.Brush.Color := FBitmap.TransparentColor;
  FBitMap.Canvas.FillRect(Rect(0, 0, FBitMap.Width, FBitMap.Height));

  // Make sure all data is loaded
  for i := 0 to High(FData) do
    if FData[i].Status = dspPending then WaitFor(i);

  // Set the dirty area in FRenderMap coordinates and make the transformation
  ViewAreaToRendererArea(clip, area);
  FTrans.Transform(FRenderMap, area, FBitmap);
end;

procedure TDSpRasterRendererSimple.SaveImage(const fn: string);
var r, c, b: integer;
    area: TRect;
    rrm: PDSpRenderRasterMap;
    pixel: TDSpFloat64;
begin
  if Length(FData) = 0 then Exit;

  if Length(FData) = 1 then begin
    New(rrm);
    r := FData[0].Band.Raster.Rows;
    c := FData[0].Band.Raster.Columns;
    FRenderMap.Resize(r, c);
    rrm^.Map := FRenderMap;
    rrm^.Min := FRenderMin;
    rrm^.Max := FRenderMax;
    rrm^.Pixel := -1; // flag full resolution
//    FData[0].RenderFullResolution(rrm);
               {
procedure TDSpRasterBand.RenderFullResolution(rrm: PDSpRenderRasterMap);
begin
  rrm^.Band := FBand;
  FMessage^.Msg := DSPMSG_RENDERRASTERMAP;
  FMessage^.Parameter := rrm;
  FStatus := dspPending;
  FRaster.Handler.QueueMessage(FMessage);
  WaitFor;
end;
           }
    Dispose(rrm);
  end else begin
    // Establish pixel size
    pixel := 0.0;
    for b := 0 to High(FData) do
      if pixel < FData[b].Band.Raster.PixelSize then pixel := FData[b].Band.Raster.PixelSize;

    // Establish r & c, based on TotalArea
    c := Ceil((FTotalArea.East - FTotalArea.West) / pixel);
    r := Ceil((FTotalArea.North - FTotalArea.South) / pixel);
    FRenderMap.Resize(r, c);
    area := Rect(0, 0, c-1, r-1);

    // Instruct the bands to fetch the data
//    for b := 0 to High(FData) do
//      FData[b].PrepareMap(FTotalArea, area, pixel, FRenderMin, FRenderMax);

    // Now render that data to the map
//    for b := 0 to High(FData) do FData[b].Render(area, FRenderMap);
  end;

  // Render the map to the bitmap
  FBitmap.TransparentColor := FTrans.Transparent;
  FBitMap.Canvas.Brush.Color := FBitmap.TransparentColor;
  FBitmap.Width := c;
  FBitmap.Height := r;
  FBitMap.Canvas.FillRect(Rect(0, 0, c, r));
//  FTrans.Transform(FRenderMap, FBitmap);

  // Save the bitmap
  SaveAsImage(FBitmap, fn);

  // Restore the viewer geometry
  TDSpViewer(FViewer).SetRendererMap;
end;

procedure TDSpRasterRendererSimple.ResetArea;
var i: integer;
begin
  if Length(FData) = 0 then Exit;
  FTotalArea.CopyFrom(FData[0].Band.Raster.Extent);
  if Length(FData) > 1 then
    for i := 1 to High(FData) do
      FTotalArea.UnionWith(FData[0].Band.Raster.Extent);
  TDSpViewer(FViewer).ResetExtent;
end;

procedure TDSpRasterRendererSimple.Range(min, max: TDSpFloat64);
begin
  FRenderMin := min;
  FRenderMax := max;
end;

procedure TDSpRasterRendererSimple.SetMin(val: TDSpFloat64);
begin
  FRenderMin := val;
end;

function TDSpRasterRendererSimple.GetMin: TDSpFloat64;
begin
  Result := FRenderMin;
end;

procedure TDSpRasterRendererSimple.SetMax(val: TDSpFloat64);
begin
  FRenderMax := val;
end;

function TDSpRasterRendererSimple.GetMax: TDSpFloat64;
begin
  Result := FRenderMax;
end;

////////////////////////////////////////////////////////////////////////////////
// Feature renderer
constructor TDSpFeatureRenderer.Create(const viewer: TObject; lab: string);
begin
  Assert(viewer is TDSpVector);
  FViewer := viewer;
  FLabel := lab;
end;

procedure TDSpFeatureRenderer.Paint;
begin
end;

procedure TDSpFeatureRenderer.SaveImage(const fn: string);
begin
end;

end.
